@@ -74,6 +74,7 @@ enum {
 	ISO_TIMESTAMP_COMMA_GT  = ISO_TIMESTAMP_COMMA_G | ISO_T
 };
 
+#define CTIME_BUFSIZ	26
 #define ISO_BUFSIZ	42
 
 int strtimeval_iso(struct timeval *tv, int flags, char *buf, size_t bufsz);
@@ -88,6 +89,4 @@ int strtime_short(const time_t *t, struct timeval *now, int flags, char *buf, si
 extern time_t timegm(struct tm *tm);
 #endif
 
-int parse_date(struct timespec *, char const *, struct timespec const *);
-
 #endif /* UTIL_LINUX_TIME_UTIL_H */
