@@ -847,6 +847,9 @@ int sysfs_devname_is_hidden(const char *prefix, const char *name)
 
 	if (strncmp("/dev/", name, 5) == 0)
 		return 0;
+
+	if (!prefix)
+		prefix = "";
 	/*
 	 * Create path to /sys/block/<name>/hidden
 	 */
@@ -898,7 +901,7 @@ dev_t __sysfs_devname_to_devno(const char *prefix, const char *name, const char
 		goto done;
 	sysfs_devname_dev_to_sys(_name);
 
-	if (parent && strncmp("dm-", name, 3)) {
+	if (parent && strncmp("dm-", name, 3) != 0) {
 		/*
 		 * Create path to /sys/block/<parent>/<name>/dev
 		 */
