@@ -27,6 +27,7 @@
 #include "strutils.h"
 #include "all-io.h"
 #include "pt-mbr.h"
+#include "encode.h"
 
 /**
  * SECTION: gpt
@@ -155,6 +156,16 @@ static struct fdisk_parttype gpt_parttypes[] =
 	#include "pt-gpt-partnames.h"
 };
 
+static const struct fdisk_shortcut gpt_parttype_cuts[] =
+{
+	{ .shortcut = "L", .alias = "linux", .data = "0FC63DAF-8483-4772-8E79-3D69D8477DE4" }, /* Linux */
+	{ .shortcut = "S", .alias = "swap",  .data = "0657FD6D-A4AB-43C4-84E5-0933C84B4F4F" }, /* Swap */
+	{ .shortcut = "H", .alias = "home",  .data = "933AC7E1-2EB4-4F13-B844-0E14E2AEF915" }, /* Home */
+	{ .shortcut = "U", .alias = "uefi",  .data = "C12A7328-F81F-11D2-BA4B-00A0C93EC93B" }, /* UEFI system */
+	{ .shortcut = "R", .alias = "raid",  .data = "A19D880F-05FC-4D3B-A006-743F0F84911E" }, /* Linux RAID */
+	{ .shortcut = "V", .alias = "lvm",   .data = "E6D6D379-F507-44C2-A23C-238F2A3DF928" }  /* LVM */
+};
+
 #define alignment_required(_x)  ((_x)->grain != (_x)->sector_size)
 
 /* gpt_entry macros */
@@ -172,6 +183,9 @@ struct fdisk_gpt_label {
 	struct gpt_header	*bheader;	/* backup header */
 
 	unsigned char *ents;			/* entries (partitions) */
+
+	unsigned int no_relocate :1,		/* do not fix backup location */
+		     minimize :1;
 };
 
 static void gpt_deinit(struct fdisk_label *lb);
@@ -221,7 +235,7 @@ static void swap_efi_guid(struct gpt_guid *uid)
 static int string_to_guid(const char *in, struct gpt_guid *guid)
 {
 	if (uuid_parse(in, (unsigned char *) guid)) {   /* BE */
-		DBG(LABEL, ul_debug("GPT: failed to parse GUID: %s", in));
+		DBG(GPT, ul_debug("failed to parse GUID: %s", in));
 		return -EINVAL;
 	}
 	swap_efi_guid(guid);				/* LE */
@@ -254,7 +268,7 @@ static struct fdisk_parttype *gpt_partition_parttype(
 static void gpt_entry_set_type(struct gpt_entry *e, struct gpt_guid *uuid)
 {
 	e->type = *uuid;
-	DBG(LABEL, gpt_debug_uuid("new type", uuid));
+	DBG(GPT, gpt_debug_uuid("new type", uuid));
 }
 
 static int gpt_entry_set_name(struct gpt_entry *e, char *str)
@@ -362,12 +376,16 @@ static inline size_t gpt_get_nentries(struct fdisk_gpt_label *gpt)
 	return (size_t) le32_to_cpu(gpt->pheader->npartition_entries);
 }
 
-static inline int gpt_calculate_sizeof_ents(struct gpt_header *hdr, uint32_t nents, size_t *sz)
+/* calculate size of entries array in bytes for specified number of entries */
+static inline int gpt_calculate_sizeof_entries(
+				struct gpt_header *hdr,
+				uint32_t nents,	size_t *sz)
 {
-	uint32_t esz = le32_to_cpu(hdr->sizeof_partition_entry);
+	uint32_t esz = hdr ? le32_to_cpu(hdr->sizeof_partition_entry) :
+			     sizeof(struct gpt_entry);
 
 	if (nents == 0 || esz == 0 || SIZE_MAX/esz < nents) {
-		DBG(LABEL, ul_debug("GPT entreis array size check failed"));
+		DBG(GPT, ul_debug("entries array size check failed"));
 		return -ERANGE;
 	}
 
@@ -375,11 +393,68 @@ static inline int gpt_calculate_sizeof_ents(struct gpt_header *hdr, uint32_t nen
 	return 0;
 }
 
-static inline int gpt_sizeof_ents(struct gpt_header *hdr, size_t *sz)
+/* calculate size of entries array in sectors for specified number of entries */
+static inline int gpt_calculate_sectorsof_entries(
+				struct gpt_header *hdr,
+				uint32_t nents, uint64_t *sz,
+				struct fdisk_context *cxt)
 {
-	return gpt_calculate_sizeof_ents(hdr, le32_to_cpu(hdr->npartition_entries), sz);
+	size_t esz = 0;
+	int rc = gpt_calculate_sizeof_entries(hdr, nents, &esz);	/* in bytes */
+
+	if (rc == 0)
+		*sz = (esz + cxt->sector_size - 1) / cxt->sector_size;
+	return rc;
 }
 
+/* calculate alternative (backup) entries array offset from primary header */
+static inline int gpt_calculate_alternative_entries_lba(
+				struct gpt_header *hdr,
+				uint32_t nents,
+				uint64_t *sz,
+				struct fdisk_context *cxt)
+{
+	uint64_t esects = 0;
+	int rc = gpt_calculate_sectorsof_entries(hdr, nents, &esects, cxt);
+
+	if (rc == 0)
+		*sz = cxt->total_sectors - 1ULL - esects;
+	return rc;
+}
+
+static inline int gpt_calculate_last_lba(
+				struct gpt_header *hdr,
+				uint32_t nents,
+				uint64_t *sz,
+				struct fdisk_context *cxt)
+{
+	uint64_t esects = 0;
+	int rc = gpt_calculate_sectorsof_entries(hdr, nents, &esects, cxt);
+
+	if (rc == 0)
+		*sz = cxt->total_sectors - 2ULL - esects;
+	return rc;
+}
+
+static inline int gpt_calculate_first_lba(
+				struct gpt_header *hdr,
+				uint32_t nents,
+				uint64_t *sz,
+				struct fdisk_context *cxt)
+{
+	uint64_t esects = 0;
+	int rc = gpt_calculate_sectorsof_entries(hdr, nents, &esects, cxt);
+
+	if (rc == 0)
+		*sz = esects + 2ULL;
+	return rc;
+}
+
+/* the current size of entries array in bytes */
+static inline int gpt_sizeof_entries(struct gpt_header *hdr, size_t *sz)
+{
+	return gpt_calculate_sizeof_entries(hdr, le32_to_cpu(hdr->npartition_entries), sz);
+}
 
 static char *gpt_get_header_id(struct gpt_header *header)
 {
@@ -426,21 +501,23 @@ static int gpt_mknew_pmbr(struct fdisk_context *cxt)
 	return 0;
 }
 
+
 /* Move backup header to the end of the device */
-static void gpt_fix_alternative_lba(struct fdisk_context *cxt, struct fdisk_gpt_label *gpt)
+static int gpt_fix_alternative_lba(struct fdisk_context *cxt, struct fdisk_gpt_label *gpt)
 {
 	struct gpt_header *p, *b;
-	uint64_t esz, esects, last;
+	uint64_t x = 0, orig;
+	size_t nents;
+	int rc;
 
 	if (!cxt)
-		return;
+		return -EINVAL;
 
 	p = gpt->pheader;	/* primary */
 	b = gpt->bheader;	/* backup */
 
-	/* count size of partitions array */
-	esz = (uint64_t) le32_to_cpu(p->npartition_entries) * sizeof(struct gpt_entry);
-	esects = (esz + cxt->sector_size - 1) / cxt->sector_size;
+	nents = le32_to_cpu(p->npartition_entries);
+	orig = le64_to_cpu(p->alternative_lba);
 
 	/* reference from primary to backup */
 	p->alternative_lba = cpu_to_le64(cxt->total_sectors - 1ULL);
@@ -450,14 +527,87 @@ static void gpt_fix_alternative_lba(struct fdisk_context *cxt, struct fdisk_gpt_
 	b->my_lba = p->alternative_lba;
 
 	/* fix backup partitions array address */
-	b->partition_entry_lba = cpu_to_le64(cxt->total_sectors - 1ULL - esects);
+	rc = gpt_calculate_alternative_entries_lba(p, nents, &x, cxt);
+	if (rc)
+		goto failed;
+
+	b->partition_entry_lba = cpu_to_le64(x);
 
 	/* update last usable LBA */
-	last = cxt->total_sectors - 2ULL - esects;
-	p->last_usable_lba  = cpu_to_le64(last);
-	b->last_usable_lba  = cpu_to_le64(last);
+	rc = gpt_calculate_last_lba(p, nents, &x, cxt);
+	if (rc)
+		goto failed;
 
-	DBG(LABEL, ul_debug("Alternative-LBA updated to: %"PRIu64, le64_to_cpu(p->alternative_lba)));
+	p->last_usable_lba  = cpu_to_le64(x);
+	b->last_usable_lba  = cpu_to_le64(x);
+
+	DBG(GPT, ul_debug("Alternative-LBA updated from %"PRIu64" to %"PRIu64,
+				orig, le64_to_cpu(p->alternative_lba)));
+	return 0;
+failed:
+	DBG(GPT, ul_debug("failed to fix alternative-LBA [rc=%d]", rc));
+	return rc;
+}
+
+static uint64_t gpt_calculate_minimal_size(struct fdisk_context *cxt, struct fdisk_gpt_label *gpt)
+{
+	size_t i;
+	uint64_t x = 0, total = 0;
+	struct gpt_header *hdr;
+
+	assert(cxt);
+	assert(gpt);
+	assert(gpt->pheader);
+	assert(gpt->ents);
+
+	hdr = gpt->pheader;
+
+	/* LBA behind the last partition */
+	for (i = 0; i < gpt_get_nentries(gpt); i++) {
+		struct gpt_entry *e = gpt_get_entry(gpt, i);
+
+		if (gpt_entry_is_used(e)) {
+			uint64_t end = gpt_partition_end(e);
+			if (end > x)
+				x = end;
+		}
+	}
+	total = x + 1;
+
+	/* the current last LBA usable for partitions */
+	gpt_calculate_last_lba(hdr, le32_to_cpu(hdr->npartition_entries), &x, cxt);
+
+	/* size of all stuff at the end of the device */
+	total += cxt->total_sectors - x;
+
+	DBG(GPT, ul_debug("minimal device is %"PRIu64, total));
+	return total;
+}
+
+static int gpt_possible_minimize(struct fdisk_context *cxt, struct fdisk_gpt_label *gpt)
+{
+	struct gpt_header *hdr = gpt->pheader;
+	uint64_t total = gpt_calculate_minimal_size(cxt, gpt);
+
+	return le64_to_cpu(hdr->alternative_lba) > (total - 1ULL);
+}
+
+/* move backup header behind the last partition */
+static int gpt_minimize_alternative_lba(struct fdisk_context *cxt, struct fdisk_gpt_label *gpt)
+{
+	uint64_t total = gpt_calculate_minimal_size(cxt, gpt);
+	uint64_t orig = cxt->total_sectors;
+	int rc;
+
+	/* Let's temporary change size of the device to recalculate backup header */
+	cxt->total_sectors = total;
+	rc = gpt_fix_alternative_lba(cxt, gpt);
+	if (rc)
+		return rc;
+
+	cxt->total_sectors = orig;
+	fdisk_label_set_changed(cxt->label, 1);
+	return 0;
 }
 
 /* some universal differences between the headers */
@@ -469,16 +619,19 @@ static void gpt_mknew_header_common(struct fdisk_context *cxt,
 
 	header->my_lba = cpu_to_le64(lba);
 
-	if (lba == GPT_PRIMARY_PARTITION_TABLE_LBA) { /* primary */
+	if (lba == GPT_PRIMARY_PARTITION_TABLE_LBA) {
+		/* primary */
 		header->alternative_lba = cpu_to_le64(cxt->total_sectors - 1ULL);
 		header->partition_entry_lba = cpu_to_le64(2ULL);
-	} else { /* backup */
-		uint64_t esz = (uint64_t) le32_to_cpu(header->npartition_entries)
-							* sizeof(struct gpt_entry);
-		uint64_t esects = (esz + cxt->sector_size - 1) / cxt->sector_size;
+
+	} else {
+		/* backup */
+		uint64_t x = 0;
+		gpt_calculate_alternative_entries_lba(header,
+				le32_to_cpu(header->npartition_entries), &x, cxt);
 
 		header->alternative_lba = cpu_to_le64(GPT_PRIMARY_PARTITION_TABLE_LBA);
-		header->partition_entry_lba = cpu_to_le64(cxt->total_sectors - 1ULL - esects);
+		header->partition_entry_lba = cpu_to_le64(x);
 	}
 }
 
@@ -584,9 +737,7 @@ static int count_first_last_lba(struct fdisk_context *cxt,
 				 uint64_t *first, uint64_t *last)
 {
 	int rc = 0;
-	uint64_t flba, llba;
-
-	uint64_t esz = 0;
+	uint64_t flba = 0, llba = 0;
 
 	assert(cxt);
 	assert(first);
@@ -595,9 +746,11 @@ static int count_first_last_lba(struct fdisk_context *cxt,
 	*first = *last = 0;
 
 	/* UEFI default */
-	esz = sizeof(struct gpt_entry) * GPT_NPARTITIONS / cxt->sector_size;
-	llba = cxt->total_sectors - 2ULL - esz;
-	flba = esz + 2ULL;
+	rc = gpt_calculate_last_lba(NULL, GPT_NPARTITIONS, &llba, cxt);
+	if (rc == 0)
+		gpt_calculate_first_lba(NULL, GPT_NPARTITIONS, &flba, cxt);
+	if (rc)
+		return rc;
 
 	/* script default */
 	if (cxt->script) {
@@ -605,7 +758,7 @@ static int count_first_last_lba(struct fdisk_context *cxt,
 		if (rc < 0)
 			return rc;
 
-		DBG(LABEL, ul_debug("FirstLBA: script=%"PRIu64", uefi=%"PRIu64", topology=%ju.",
+		DBG(GPT, ul_debug("FirstLBA: script=%"PRIu64", uefi=%"PRIu64", topology=%ju.",
 		                    *first, flba,  (uintmax_t)cxt->first_lba));
 
 		if (rc == 0 && (*first < flba || *first > llba)) {
@@ -617,7 +770,7 @@ static int count_first_last_lba(struct fdisk_context *cxt,
 		if (rc < 0)
 			return rc;
 
-		DBG(LABEL, ul_debug("LastLBA: script=%"PRIu64", uefi=%"PRIu64", topology=%ju.",
+		DBG(GPT, ul_debug("LastLBA: script=%"PRIu64", uefi=%"PRIu64", topology=%ju.",
 		                    *last, llba,  (uintmax_t)cxt->last_lba));
 
 		if (rc == 0 && (*last > llba || *last < flba)) {
@@ -775,7 +928,7 @@ static int valid_pmbr(struct fdisk_context *cxt)
 		}
 	}
 done:
-	DBG(LABEL, ul_debug("PMBR type: %s",
+	DBG(GPT, ul_debug("PMBR type: %s",
 			ret == GPT_MBR_PROTECTIVE ? "protective" :
 			ret == GPT_MBR_HYBRID     ? "hybrid"     : "???" ));
 	return ret;
@@ -800,7 +953,7 @@ static uint64_t last_lba(struct fdisk_context *cxt)
 	else
 		fdisk_warnx(cxt, _("gpt: cannot handle files with mode %o"), s.st_mode);
 
-	DBG(LABEL, ul_debug("GPT last LBA: %"PRIu64"", sectors));
+	DBG(GPT, ul_debug("last LBA: %"PRIu64"", sectors));
 	return sectors;
 }
 
@@ -828,11 +981,11 @@ static unsigned char *gpt_read_entries(struct fdisk_context *cxt,
 	assert(cxt);
 	assert(header);
 
-	if (gpt_sizeof_ents(header, &sz))
+	if (gpt_sizeof_entries(header, &sz))
 		return NULL;
 
 	if (sz > (size_t) SSIZE_MAX) {
-		DBG(LABEL, ul_debug("GPT entries array too large to read()"));
+		DBG(GPT, ul_debug("entries array too large to read()"));
 		return NULL;
 	}
 
@@ -875,7 +1028,7 @@ static inline uint32_t gpt_entryarr_count_crc32(struct gpt_header *header, unsig
 {
 	size_t arysz = 0;
 
-	if (gpt_sizeof_ents(header, &arysz))
+	if (gpt_sizeof_entries(header, &arysz))
 		return 0;
 
 	return count_crc32(ents, arysz, 0, 0);
@@ -946,20 +1099,20 @@ static int gpt_check_lba_sanity(struct fdisk_context *cxt, struct gpt_header *he
 
 	/* check if first and last usable LBA make sense */
 	if (lu < fu) {
-		DBG(LABEL, ul_debug("error: header last LBA is before first LBA"));
+		DBG(GPT, ul_debug("error: header last LBA is before first LBA"));
 		goto done;
 	}
 
 	/* check if first and last usable LBAs with the disk's last LBA */
 	if (fu > lastlba || lu > lastlba) {
-		DBG(LABEL, ul_debug("error: header LBAs are after the disk's last LBA"));
+		DBG(GPT, ul_debug("error: header LBAs are after the disk's last LBA"));
 		goto done;
 	}
 
 	/* the header has to be outside usable range */
 	if (fu < GPT_PRIMARY_PARTITION_TABLE_LBA &&
 	    GPT_PRIMARY_PARTITION_TABLE_LBA < lu) {
-		DBG(LABEL, ul_debug("error: header outside of usable range"));
+		DBG(GPT, ul_debug("error: header outside of usable range"));
 		goto done;
 	}
 
@@ -1033,13 +1186,13 @@ static struct gpt_header *gpt_read_header(struct fdisk_context *cxt,
 	else
 		free(ents);
 
-	DBG(LABEL, ul_debug("found valid GPT Header on LBA %"PRIu64"", lba));
+	DBG(GPT, ul_debug("found valid header on LBA %"PRIu64"", lba));
 	return header;
 invalid:
 	free(header);
 	free(ents);
 
-	DBG(LABEL, ul_debug("read GPT Header on LBA %"PRIu64" failed", lba));
+	DBG(GPT, ul_debug("read header on LBA %"PRIu64" failed", lba));
 	return NULL;
 }
 
@@ -1071,7 +1224,7 @@ static int gpt_locate_disklabel(struct fdisk_context *cxt, int n,
 		gpt = self_label(cxt);
 		*offset = (uint64_t) le64_to_cpu(gpt->pheader->partition_entry_lba) *
 				     cxt->sector_size;
-		return gpt_sizeof_ents(gpt->pheader, size);
+		return gpt_sizeof_entries(gpt->pheader, size);
 	default:
 		return 1;			/* no more chunks */
 	}
@@ -1237,7 +1390,7 @@ static uint32_t check_overlap_partitions(struct fdisk_gpt_label *gpt)
 			if (!gpt_entry_is_used(ei) || !gpt_entry_is_used(ej))
 				continue;
 			if (partition_overlap(ei, ej)) {
-				DBG(LABEL, ul_debug("GPT partitions overlap detected [%zu vs. %zu]", i, j));
+				DBG(GPT, ul_debug("partitions overlap detected [%zu vs. %zu]", i, j));
 				return i + 1;
 			}
 		}
@@ -1484,65 +1637,48 @@ static int gpt_probe_label(struct fdisk_context *cxt)
 
 	/* The headers make be correct, but Backup do not have to be on the end
 	 * of the device (due to device resize, etc.). Let's fix this issue. */
-	if (le64_to_cpu(gpt->pheader->alternative_lba) > cxt->total_sectors ||
-	    le64_to_cpu(gpt->pheader->alternative_lba) < cxt->total_sectors - 1ULL) {
-		fdisk_warnx(cxt, _("The backup GPT table is not on the end of the device. "
-				   "This problem will be corrected by write."));
+	if (gpt->minimize == 0 &&
+	    (le64_to_cpu(gpt->pheader->alternative_lba) > cxt->total_sectors ||
+	     le64_to_cpu(gpt->pheader->alternative_lba) < cxt->total_sectors - 1ULL)) {
 
-		gpt_fix_alternative_lba(cxt, gpt);
-		gpt_recompute_crc(gpt->bheader, gpt->ents);
-		gpt_recompute_crc(gpt->pheader, gpt->ents);
-		fdisk_label_set_changed(cxt->label, 1);
+		if (gpt->no_relocate || fdisk_is_readonly(cxt))
+			fdisk_warnx(cxt, _("The backup GPT table is not on the end of the device."));
+
+		else {
+			fdisk_warnx(cxt, _("The backup GPT table is not on the end of the device. "
+					   "This problem will be corrected by write."));
+
+			if (gpt_fix_alternative_lba(cxt, gpt) != 0)
+				fdisk_warnx(cxt, _("Failed to recalculate backup GPT table location"));
+			gpt_recompute_crc(gpt->bheader, gpt->ents);
+			gpt_recompute_crc(gpt->pheader, gpt->ents);
+			fdisk_label_set_changed(cxt->label, 1);
+		}
 	}
 
+	if (gpt->minimize && gpt_possible_minimize(cxt, gpt))
+		fdisk_label_set_changed(cxt->label, 1);
+
 	cxt->label->nparts_max = gpt_get_nentries(gpt);
 	cxt->label->nparts_cur = partitions_in_use(gpt);
 	return 1;
 failed:
-	DBG(LABEL, ul_debug("GPT probe failed"));
+	DBG(GPT, ul_debug("probe failed"));
 	gpt_deinit(cxt->label);
 	return 0;
 }
 
-/*
- * Stolen from libblkid - can be removed once partition semantics
- * are added to the fdisk API.
- */
 static char *encode_to_utf8(unsigned char *src, size_t count)
 {
-	uint16_t c;
-	char *dest;
-	size_t i, j;
-	size_t len = count * 3 / 2;
+	unsigned char *dest;
+	size_t len = (count * 3 / 2) + 1;
 
-	dest = calloc(1, len + 1);
+	dest = calloc(1, len);
 	if (!dest)
 		return NULL;
 
-	for (j = i = 0; i + 2 <= count; i += 2) {
-		/* always little endian */
-		c = (src[i+1] << 8) | src[i];
-		if (c == 0) {
-			break;
-		} else if (c < 0x80) {
-			if (j+1 > len)
-				break;
-			dest[j++] = (uint8_t) c;
-		} else if (c < 0x800) {
-			if (j+2 > len)
-				break;
-			dest[j++] = (uint8_t) (0xc0 | (c >> 6));
-			dest[j++] = (uint8_t) (0x80 | (c & 0x3f));
-		} else {
-			if (j+3 > len)
-				break;
-			dest[j++] = (uint8_t) (0xe0 | (c >> 12));
-			dest[j++] = (uint8_t) (0x80 | ((c >> 6) & 0x3f));
-			dest[j++] = (uint8_t) (0x80 | (c & 0x3f));
-		}
-	}
-
-	return dest;
+	ul_encode_to_utf8(UL_ENCODE_UTF16LE, dest, len, src, count);
+	return (char *) dest;
 }
 
 static int gpt_entry_attrs_to_string(struct gpt_entry *e, char **res)
@@ -1577,13 +1713,13 @@ static int gpt_entry_attrs_to_string(struct gpt_entry *e, char **res)
 		p += l - 1;
 	}
 	if (isset(bits, GPT_ATTRBIT_NOBLOCK)) {
-		if (p > *res)
+		if (p != *res)
 			*p++ = ' ';
 		memcpy(p, GPT_ATTRSTR_NOBLOCK, (l = sizeof(GPT_ATTRSTR_NOBLOCK)));
 		p += l - 1;
 	}
 	if (isset(bits, GPT_ATTRBIT_LEGACY)) {
-		if (p > *res)
+		if (p != *res)
 			*p++ = ' ';
 		memcpy(p, GPT_ATTRSTR_LEGACY, (l = sizeof(GPT_ATTRSTR_LEGACY)));
 		p += l - 1;
@@ -1595,7 +1731,7 @@ static int gpt_entry_attrs_to_string(struct gpt_entry *e, char **res)
 		if (!isset(bits, n))
 			continue;
 		if (!count) {
-			if (p > *res)
+			if (p != *res)
 				*p++ = ' ';
 			p += sprintf(p, "GUID:%u", n);
 		} else
@@ -1618,7 +1754,7 @@ static int gpt_entry_attrs_from_string(
 	assert(e);
 	assert(p);
 
-	DBG(LABEL, ul_debug("GPT: parsing string attributes '%s'", p));
+	DBG(GPT, ul_debug("parsing string attributes '%s'", p));
 
 	bits = (char *) &attrs;
 
@@ -1629,7 +1765,7 @@ static int gpt_entry_attrs_from_string(
 		if (!*p)
 			break;
 
-		DBG(LABEL, ul_debug(" parsing item '%s'", p));
+		DBG(GPT, ul_debug(" item '%s'", p));
 
 		if (strncmp(p, GPT_ATTRSTR_REQ,
 					sizeof(GPT_ATTRSTR_REQ) - 1) == 0) {
@@ -1832,7 +1968,20 @@ static int gpt_set_partition(struct fdisk_context *cxt, size_t n,
 	return rc;
 }
 
+static int gpt_write(struct fdisk_context *cxt, off_t offset, void *buf, size_t count)
+{
+	if (offset != lseek(cxt->dev_fd, offset, SEEK_SET))
+		return -errno;
 
+	if (write_all(cxt->dev_fd, buf, count))
+		return -errno;
+
+	fsync(cxt->dev_fd);
+
+	DBG(GPT, ul_debug("  write OK [offset=%zu, size=%zu]",
+				(size_t) offset, count));
+	return 0;
+}
 
 /*
  * Write partitions.
@@ -1841,23 +1990,16 @@ static int gpt_set_partition(struct fdisk_context *cxt, size_t n,
 static int gpt_write_partitions(struct fdisk_context *cxt,
 				struct gpt_header *header, unsigned char *ents)
 {
-	off_t offset = (off_t) le64_to_cpu(header->partition_entry_lba) * cxt->sector_size;
-	size_t towrite = 0;
-	ssize_t ssz;
+	size_t esz = 0;
 	int rc;
 
-	rc = gpt_sizeof_ents(header, &towrite);
+	rc = gpt_sizeof_entries(header, &esz);
 	if (rc)
 		return rc;
 
-	if (offset != lseek(cxt->dev_fd, offset, SEEK_SET))
-		return -errno;
-
-	ssz = write(cxt->dev_fd, ents, towrite);
-	if (ssz < 0 || (ssize_t) towrite != ssz)
-		return -errno;
-
-	return 0;
+	return gpt_write(cxt,
+			(off_t) le64_to_cpu(header->partition_entry_lba) * cxt->sector_size,
+			ents, esz);
 }
 
 /*
@@ -1871,15 +2013,7 @@ static int gpt_write_partitions(struct fdisk_context *cxt,
 static int gpt_write_header(struct fdisk_context *cxt,
 			    struct gpt_header *header, uint64_t lba)
 {
-	off_t offset = lba * cxt->sector_size;
-
-	if (offset != lseek(cxt->dev_fd, offset, SEEK_SET))
-		goto fail;
-	if (cxt->sector_size ==
-	    (size_t) write(cxt->dev_fd, header, cxt->sector_size))
-		return 0;
-fail:
-	return -errno;
+	return gpt_write(cxt, lba * cxt->sector_size, header, cxt->sector_size);
 }
 
 /*
@@ -1888,13 +2022,12 @@ fail:
  */
 static int gpt_write_pmbr(struct fdisk_context *cxt)
 {
-	off_t offset;
 	struct gpt_legacy_mbr *pmbr;
 
 	assert(cxt);
 	assert(cxt->firstsector);
 
-	DBG(LABEL, ul_debug("(over)writing PMBR"));
+	DBG(GPT, ul_debug("(over)writing PMBR"));
 	pmbr = (struct gpt_legacy_mbr *) cxt->firstsector;
 
 	/* zero out the legacy partitions */
@@ -1918,16 +2051,9 @@ static int gpt_write_pmbr(struct fdisk_context *cxt)
 		pmbr->partition_record[0].size_in_lba =
 			cpu_to_le32((uint32_t) (cxt->total_sectors - 1ULL));
 
-	offset = GPT_PMBR_LBA * cxt->sector_size;
-	if (offset != lseek(cxt->dev_fd, offset, SEEK_SET))
-		goto fail;
-
 	/* pMBR covers the first sector (LBA) of the disk */
-	if (write_all(cxt->dev_fd, pmbr, cxt->sector_size))
-		goto fail;
-	return 0;
-fail:
-	return -errno;
+	return gpt_write(cxt, GPT_PMBR_LBA * cxt->sector_size,
+			 pmbr, cxt->sector_size);
 }
 
 /*
@@ -1944,6 +2070,8 @@ static int gpt_write_disklabel(struct fdisk_context *cxt)
 	assert(cxt->label);
 	assert(fdisk_is_label(cxt, GPT));
 
+	DBG(GPT, ul_debug("writing..."));
+
 	gpt = self_label(cxt);
 	mbr_type = valid_pmbr(cxt);
 
@@ -1958,6 +2086,9 @@ static int gpt_write_disklabel(struct fdisk_context *cxt)
 	if (check_overlap_partitions(gpt))
 		goto err0;
 
+	if (gpt->minimize)
+		gpt_minimize_alternative_lba(cxt, gpt);
+
 	/* recompute CRCs for both headers */
 	gpt_recompute_crc(gpt->pheader, gpt->ents);
 	gpt_recompute_crc(gpt->bheader, gpt->ents);
@@ -1987,14 +2118,14 @@ static int gpt_write_disklabel(struct fdisk_context *cxt)
 	else if (gpt_write_pmbr(cxt) != 0)
 		goto err1;
 
-	DBG(LABEL, ul_debug("GPT write success"));
+	DBG(GPT, ul_debug("...write success"));
 	return 0;
 err0:
-	DBG(LABEL, ul_debug("GPT write failed: incorrect input"));
+	DBG(GPT, ul_debug("...write failed: incorrect input"));
 	errno = EINVAL;
 	return -EINVAL;
 err1:
-	DBG(LABEL, ul_debug("GPT write failed: %m"));
+	DBG(GPT, ul_debug("...write failed: %m"));
 	return -errno;
 }
 
@@ -2118,7 +2249,7 @@ static int gpt_verify_disklabel(struct fdisk_context *cxt)
 			P_("%d error detected.", "%d errors detected.", nerror),
 			nerror);
 
-	return 0;
+	return nerror;
 }
 
 /* Delete a single GPT partition, specified by partnum. */
@@ -2182,7 +2313,7 @@ static int gpt_add_partition(
 
 	rc = fdisk_partition_next_partno(pa, cxt, &partnum);
 	if (rc) {
-		DBG(LABEL, ul_debug("GPT failed to get next partno"));
+		DBG(GPT, ul_debug("failed to get next partno"));
 		return rc;
 	}
 
@@ -2219,14 +2350,14 @@ static int gpt_add_partition(
 
 		do {
 			uint64_t x;
-			DBG(LABEL, ul_debug("testing first sector %"PRIu64"", disk_f));
+			DBG(GPT, ul_debug("testing first sector %"PRIu64"", disk_f));
 			disk_f = find_first_available(gpt, disk_f);
 			if (!disk_f)
 				break;
 			x = find_last_free(gpt, disk_f);
 			if (x - disk_f >= cxt->grain / cxt->sector_size)
 				break;
-			DBG(LABEL, ul_debug("first sector %"PRIu64" addresses to small space, continue...", disk_f));
+			DBG(GPT, ul_debug("first sector %"PRIu64" addresses to small space, continue...", disk_f));
 			disk_f = x + 1ULL;
 		} while(1);
 
@@ -2249,7 +2380,7 @@ static int gpt_add_partition(
 		user_f = dflt_f;
 
 	} else if (pa && fdisk_partition_has_start(pa)) {
-		DBG(LABEL, ul_debug("first sector defined: %ju",  (uintmax_t)pa->start));
+		DBG(GPT, ul_debug("first sector defined: %ju",  (uintmax_t)pa->start));
 		if (pa->start != find_first_available(gpt, pa->start)) {
 			fdisk_warnx(cxt, _("Sector %ju already used."),  (uintmax_t)pa->start);
 			return -ERANGE;
@@ -2294,7 +2425,7 @@ static int gpt_add_partition(
 
 	} else if (pa && fdisk_partition_has_size(pa)) {
 		user_l = user_f + pa->size - 1;
-		DBG(LABEL, ul_debug("size defined: %ju, end: %"PRIu64" (last possible: %"PRIu64")",
+		DBG(GPT, ul_debug("size defined: %ju, end: %"PRIu64" (last possible: %"PRIu64")",
 					 (uintmax_t)pa->size, user_l, dflt_l));
 
 		if (user_l != dflt_l
@@ -2398,7 +2529,7 @@ static int gpt_add_partition(
 	if (pa && pa->attrs)
 		gpt_entry_attrs_from_string(cxt, e, pa->attrs);
 
-	DBG(LABEL, ul_debug("GPT new partition: partno=%zu, start=%"PRIu64", end=%"PRIu64", size=%"PRIu64"",
+	DBG(GPT, ul_debug("new partition: partno=%zu, start=%"PRIu64", end=%"PRIu64", size=%"PRIu64"",
 				partnum,
 				gpt_partition_start(e),
 				gpt_partition_end(e),
@@ -2480,7 +2611,7 @@ static int gpt_create_disklabel(struct fdisk_context *cxt)
 	if (rc < 0)
 		goto done;
 
-	rc = gpt_sizeof_ents(gpt->pheader, &esz);
+	rc = gpt_sizeof_entries(gpt->pheader, &esz);
 	if (rc)
 		goto done;
 	gpt->ents = calloc(1, esz);
@@ -2502,11 +2633,11 @@ done:
 	return rc;
 }
 
-static int gpt_set_disklabel_id(struct fdisk_context *cxt)
+static int gpt_set_disklabel_id(struct fdisk_context *cxt, const char *str)
 {
 	struct fdisk_gpt_label *gpt;
 	struct gpt_guid uuid;
-	char *str, *old, *new;
+	char *old, *new;
 	int rc;
 
 	assert(cxt);
@@ -2514,12 +2645,16 @@ static int gpt_set_disklabel_id(struct fdisk_context *cxt)
 	assert(fdisk_is_label(cxt, GPT));
 
 	gpt = self_label(cxt);
-	if (fdisk_ask_string(cxt,
-			_("Enter new disk UUID (in 8-4-4-4-12 format)"), &str))
-		return -EINVAL;
+	if (!str) {
+		char *buf = NULL;
 
-	rc = string_to_guid(str, &uuid);
-	free(str);
+		if (fdisk_ask_string(cxt,
+				_("Enter new disk UUID (in 8-4-4-4-12 format)"), &buf))
+			return -EINVAL;
+		rc = string_to_guid(buf, &uuid);
+		free(buf);
+	} else
+		rc = string_to_guid(str, &uuid);
 
 	if (rc) {
 		fdisk_warnx(cxt, _("Failed to parse your UUID."));
@@ -2583,7 +2718,7 @@ static int gpt_check_table_overlap(struct fdisk_context *cxt,
 /**
  * fdisk_gpt_set_npartitions:
  * @cxt: context
- * @entries: new size
+ * @nents: number of wanted entries
  *
  * Elarge GPT entries array if possible. The function check if an existing
  * partition does not overlap the entries array area. If yes, then it report
@@ -2592,12 +2727,12 @@ static int gpt_check_table_overlap(struct fdisk_context *cxt,
  * Returns: 0 on success, < 0 on error.
  * Since: 2.29
  */
-int fdisk_gpt_set_npartitions(struct fdisk_context *cxt, uint32_t entries)
+int fdisk_gpt_set_npartitions(struct fdisk_context *cxt, uint32_t nents)
 {
 	struct fdisk_gpt_label *gpt;
-	size_t old_size, new_size;
-	uint32_t old;
-	uint64_t first_usable, last_usable;
+	size_t new_size = 0;
+	uint32_t old_nents;
+	uint64_t first_usable = 0ULL, last_usable = 0ULL;
 	int rc;
 
 	assert(cxt);
@@ -2608,12 +2743,12 @@ int fdisk_gpt_set_npartitions(struct fdisk_context *cxt, uint32_t entries)
 
 	gpt = self_label(cxt);
 
-	old = le32_to_cpu(gpt->pheader->npartition_entries);
-	if (old == entries)
+	old_nents = le32_to_cpu(gpt->pheader->npartition_entries);
+	if (old_nents == nents)
 		return 0;	/* do nothing, say nothing */
 
 	/* calculate the size (bytes) of the entries array */
-	rc = gpt_calculate_sizeof_ents(gpt->pheader, entries, &new_size);
+	rc = gpt_calculate_sizeof_entries(gpt->pheader, nents, &new_size);
 	if (rc) {
 		uint32_t entry_size = le32_to_cpu(gpt->pheader->sizeof_partition_entry);
 
@@ -2625,20 +2760,21 @@ int fdisk_gpt_set_npartitions(struct fdisk_context *cxt, uint32_t entries)
 		return rc;
 	}
 
-	rc = gpt_calculate_sizeof_ents(gpt->pheader, old, &old_size);
+	rc = gpt_calculate_first_lba(gpt->pheader, nents, &first_usable, cxt);
+	if (rc == 0)
+		rc = gpt_calculate_last_lba(gpt->pheader, nents, &last_usable, cxt);
 	if (rc)
 		return rc;
 
-	/* calculate new range of usable LBAs */
-	first_usable = (uint64_t) (new_size / cxt->sector_size) + 2ULL;
-	last_usable = cxt->total_sectors - 2ULL - (uint64_t) (new_size / cxt->sector_size);
-
 	/* if expanding the table, first check that everything fits,
 	 * then allocate more memory and zero. */
-	if (entries > old) {
+	if (nents > old_nents) {
 		unsigned char *ents;
+		size_t old_size = 0;
 
-		rc = gpt_check_table_overlap(cxt, first_usable, last_usable);
+		rc = gpt_calculate_sizeof_entries(gpt->pheader, old_nents, &old_size);
+		if (rc == 0)
+			rc = gpt_check_table_overlap(cxt, first_usable, last_usable);
 		if (rc)
 			return rc;
 		ents = realloc(gpt->ents, new_size);
@@ -2651,8 +2787,8 @@ int fdisk_gpt_set_npartitions(struct fdisk_context *cxt, uint32_t entries)
 	}
 
 	/* everything's ok, apply the new size */
-	gpt->pheader->npartition_entries = cpu_to_le32(entries);
-	gpt->bheader->npartition_entries = cpu_to_le32(entries);
+	gpt->pheader->npartition_entries = cpu_to_le32(nents);
+	gpt->bheader->npartition_entries = cpu_to_le32(nents);
 
 	/* usable LBA addresses will have changed */
 	fdisk_set_first_lba(cxt, first_usable);
@@ -2662,7 +2798,6 @@ int fdisk_gpt_set_npartitions(struct fdisk_context *cxt, uint32_t entries)
 	gpt->pheader->last_usable_lba = cpu_to_le64(last_usable);
 	gpt->bheader->last_usable_lba = cpu_to_le64(last_usable);
 
-
 	/* The backup header must be recalculated */
 	gpt_mknew_header_common(cxt, gpt->bheader, le64_to_cpu(gpt->pheader->alternative_lba));
 
@@ -2673,7 +2808,8 @@ int fdisk_gpt_set_npartitions(struct fdisk_context *cxt, uint32_t entries)
 	/* update library info */
 	cxt->label->nparts_max = gpt_get_nentries(gpt);
 
-	fdisk_info(cxt, _("Partition table length changed from %"PRIu32" to %"PRIu64"."), old, entries);
+	fdisk_info(cxt, _("Partition table length changed from %"PRIu32" to %"PRIu64"."),
+			old_nents, nents);
 
 	fdisk_label_set_changed(cxt->label, 1);
 	return 0;
@@ -2775,7 +2911,7 @@ int fdisk_gpt_set_partition_attrs(
 	if (!fdisk_is_label(cxt, GPT))
 		return -EINVAL;
 
-	DBG(LABEL, ul_debug("GPT entry attributes change requested partno=%zu", partnum));
+	DBG(GPT, ul_debug("entry attributes change requested partno=%zu", partnum));
 	gpt = self_label(cxt);
 
 	if (partnum >= gpt_get_nentries(gpt))
@@ -2808,7 +2944,7 @@ static int gpt_toggle_partition_flag(
 	assert(cxt->label);
 	assert(fdisk_is_label(cxt, GPT));
 
-	DBG(LABEL, ul_debug("GPT entry attribute change requested partno=%zu", i));
+	DBG(GPT, ul_debug("entry attribute change requested partno=%zu", i));
 	gpt = self_label(cxt);
 
 	if (i >= gpt_get_nentries(gpt))
@@ -3031,8 +3167,11 @@ struct fdisk_label *fdisk_new_gpt_label(struct fdisk_context *cxt __attribute__
 	lb->name = "gpt";
 	lb->id = FDISK_DISKLABEL_GPT;
 	lb->op = &gpt_operations;
+
 	lb->parttypes = gpt_parttypes;
 	lb->nparttypes = ARRAY_SIZE(gpt_parttypes);
+	lb->parttype_cuts = gpt_parttype_cuts;
+	lb->nparttype_cuts = ARRAY_SIZE(gpt_parttype_cuts);
 
 	lb->fields = gpt_fields;
 	lb->nfields = ARRAY_SIZE(gpt_fields);
@@ -3040,6 +3179,44 @@ struct fdisk_label *fdisk_new_gpt_label(struct fdisk_context *cxt __attribute__
 	return lb;
 }
 
+/**
+ * fdisk_gpt_disable_relocation
+ * @lb: label
+ * @disable: 0 or 1
+ *
+ * Disable automatic backup header relocation to the end of the device. The
+ * header position is recalculated during libfdisk probing stage by
+ * fdisk_assign_device() and later written by fdisk_write_disklabel(), so you
+ * need to call it before fdisk_assign_device().
+ *
+ * Since: 2.36
+ */
+void fdisk_gpt_disable_relocation(struct fdisk_label *lb, int disable)
+{
+	struct fdisk_gpt_label *gpt = (struct fdisk_gpt_label *) lb;
+
+	assert(gpt);
+	gpt->no_relocate = disable ? 1 : 0;
+}
+
+/**
+ * fdisk_gpt_enable_minimize
+ * @lb: label
+ * @enable: 0 or 1
+ *
+ * Force libfdisk to write backup header to behind last partition. The
+ * header position is recalculated on fdisk_write_disklabel().
+ *
+ * Since: 2.36
+ */
+void fdisk_gpt_enable_minimize(struct fdisk_label *lb, int enable)
+{
+	struct fdisk_gpt_label *gpt = (struct fdisk_gpt_label *) lb;
+
+	assert(gpt);
+	gpt->minimize = enable ? 1 : 0;
+}
+
 #ifdef TEST_PROGRAM
 static int test_getattr(struct fdisk_test *ts, int argc, char *argv[])
 {
